function y = BjerkPrice1993(S,K,T,r,b,sigma,PutCall)

% Bjerksund and Stensland flat boundary call and put approximation

if strcmp(PutCall,'C')
    % Equation (4)
    y = CallPrice1993(S,K,T,r,b,sigma);
elseif strcmp(PutCall,'P')
    % Equation (19)
    y = CallPrice1993(K,S,T,r-b,-b,sigma);
end

function Call = CallPrice1993(S,K,T,r,b,sigma)

% Bjerksund and Stensland (1993)) flat boundary call approximation

% Black Scholes
BSCall = @(S,K,r,b,v,T) S*exp((b-r)*T)*normcdf((log(S/K) + (b+v^2/2)*T)/(v*sqrt(T))) - K*exp(-r*T)*normcdf((log(S/K) + (b+v^2/2)*T)/(v*sqrt(T)) - v*sqrt(T));

% If b >= r then it's not optimal to exercise before expiry
if b >= r
    Call = BSCall(S,K,r,b,sigma,T);
else
    % Equation (6)
    beta = (0.5 - b/sigma^2) + sqrt((b/sigma^2 - 0.5)^2 + 2*r/sigma^2);

    % Equation (10) through (13)
    Binf = K*beta/(beta-1);
    B0 = max(K,K*r/(r-b));
    h = -(b*T + 2*sigma*sqrt(T)) * K^2/B0/(Binf-B0);
    X = B0 + (Binf-B0)*(1-exp(h));

    % Determine boundary condition
    if (S >= X)
        Call = S - K;
    else
        % Equation (5)
        alpha = (X - K)*X^(-beta);

        % Call price Equation (4)
        Call = alpha*S^beta ...
            - alpha*Phi(S,T,beta,X,X,r,b,sigma) ...
            + Phi(S,T,1,X,X,r,b,sigma) ...
            - Phi(S,T,1,K,X,r,b,sigma) ...
            - K*Phi(S,T,0,X,X,r,b,sigma) ...
            + K*Phi(S,T,0,K,X,r,b,sigma);
    end
end

function BSphi = Phi(S,T,gamma,H,X,r,b,sigma)

% Bjerksund and Stensland "phi" function

% Equation 8
lambda = -r + gamma*b + 0.5*gamma*(gamma-1)*sigma^2;
% Equation 9
kappa = 2*b/sigma^2 + (2*gamma-1);

% Equation 7 for the phi function
d1 = -(log(S/H)     + (b+(gamma-0.5)*sigma^2)*T)/sigma/sqrt(T);
d2 = -(log(X^2/S/H) + (b+(gamma-0.5)*sigma^2)*T)/sigma/sqrt(T);

% Return phi
BSphi = exp(lambda*T)*S^gamma * (normcdf(d1) - (X/S)^kappa*normcdf(d2));

