function y = BjerkPrice2002(S,K,T,r,b,sigma,PutCall)

% Bjerksund and Stensland two-step boundary call and put approximation

if strcmp(PutCall,'C')
    % Equation (4)
    y = CallPrice2002(S,K,T,r,b,sigma);
elseif strcmp(PutCall,'P')
    % Equation (19)
    y = CallPrice2002(K,S,T,r-b,-b,sigma);
end


function Call = CallPrice2002(S,K,T,r,b,sigma)

% Bjerksund and Stensland two-step boundary call approximation

% Black Scholes
BSCall = @(S,K,r,b,v,T) S*exp((b-r)*T)*normcdf((log(S/K) + (b+v^2/2)*T)/(v*sqrt(T))) - K*exp(-r*T)*normcdf((log(S/K) + (b+v^2/2)*T)/(v*sqrt(T)) - v*sqrt(T));

% If b >= r then it's not optimal to exercise before expiry
if b >= r
    Call = BSCall(S,K,r,b,sigma,T);
else
    % Equation (6)
    beta = (0.5 - b/sigma^2) + sqrt((b/sigma^2 - 0.5)^2 + 2*r/sigma^2);

    % Equation (10) through (13)
    Binf = K*beta/(beta-1);
    B0 = max(K,K*r/(r-b));
    hT = -(b*T + 2*sigma*sqrt(T)) * K^2/B0/(Binf-B0);
    X = B0 + (Binf-B0)*(1-exp(hT));

    % Determine boundary condition
    if (S >= X)
        Call = S - K;
    else
        % Equation (5) alpha(X)
        alphaX = (X - K)*X^(-beta);

        % Equation (16)
        t = 0.5*(sqrt(5) - 1)*T;

        % Equation (18)
        hTt = -(b*(T-t) + 2*sigma*sqrt(T-t)) * K^2/B0/(Binf-B0);

        % Redefine x and alpha(x) in terms of t
        x = B0 + (Binf-B0)*(1-exp(hTt));
        alphax = (x - K)*x^(-beta);

        % Call price Equation (4)
        Call = alphaX*S^beta ...
            - alphaX*Phi(S,t,beta,X,X,r,b,sigma) ...
            +       Phi(S,t,1,X,X,r,b,sigma) ...
            -       Phi(S,t,1,x,X,r,b,sigma) ...
            - K*    Phi(S,t,0,X,X,r,b,sigma) ...
            + K*    Phi(S,t,0,x,X,r,b,sigma) ...
            + alphax*Phi(S,t,beta,x,X,r,b,sigma) ...
            - alphax*Psi(S,T,beta,x,X,x,t,r,b,sigma) ...
            +       Psi(S,T,1,   x,X,x,t,r,b,sigma) ...
            -       Psi(S,T,1,   K,X,x,t,r,b,sigma) ...
            - K*    Psi(S,T,0,   x,X,x,t,r,b,sigma) ...
            + K*    Psi(S,T,0,   K,X,x,t,r,b,sigma);
    end
end

function BSphi = Phi(S,T,gamma,H,X,r,b,sigma)

% Bjerksund and Stensland "phi" function

% Equation 8
lambda = -r + gamma*b + 0.5*gamma*(gamma-1)*sigma^2;
% Equation 9
kappa = 2*b/sigma^2 + (2*gamma-1);

% Equation 7 for the phi function
d1 = -(log(S/H)     + (b+(gamma-0.5)*sigma^2)*T)/sigma/sqrt(T);
d2 = -(log(X^2/S/H) + (b+(gamma-0.5)*sigma^2)*T)/sigma/sqrt(T);

% Return phi
BSphi = exp(lambda*T)*S^gamma * (normcdf(d1) - (X/S)^kappa*normcdf(d2));



function BSpsi = Psi(S,T,gamma,H,X,x,t,r,b,sigma)

% Bjerksund and Stensland "psi" function

d1 = -(log(S/x)      + (b+(gamma-0.5)*sigma^2)*t)/sigma/sqrt(t);
d2 = -(log(X^2/S/x)  + (b+(gamma-0.5)*sigma^2)*t)/sigma/sqrt(t);
d3 = -(log(S/x)      - (b+(gamma-0.5)*sigma^2)*t)/sigma/sqrt(t);
d4 = -(log(X^2/S/x)  - (b+(gamma-0.5)*sigma^2)*t)/sigma/sqrt(t);

D1 = -(log(S/H)         + (b+(gamma-0.5)*sigma^2)*T)/sigma/sqrt(T);
D2 = -(log(X^2/S/H)     + (b+(gamma-0.5)*sigma^2)*T)/sigma/sqrt(T);
D3 = -(log(x^2/S/H)     + (b+(gamma-0.5)*sigma^2)*T)/sigma/sqrt(T);
D4 = -(log(S*x^2/H/X^2) + (b+(gamma-0.5)*sigma^2)*T)/sigma/sqrt(T);

% Correlation matrix
rho = sqrt(t/T);
rhopos = [1  rho;  rho  1];
rhoneg = [1 -rho; -rho  1];

% Equation 8
lambda = -r + gamma*b + 0.5*gamma*(gamma-1)*sigma^2;
% Equation 9
kappa = 2*b/sigma^2 + (2*gamma-1);

% Return psi on page 5
BSpsi = exp(lambda*T)*S^gamma*(...
    mvncdf([d1; D1],[0;0], rhopos) ...
    - mvncdf([d2; D2],[0;0], rhopos)*(X/S)^kappa ...
    - mvncdf([d3; D3],[0;0], rhoneg)*(x/S)^kappa ...
    + mvncdf([d4; D4],[0;0], rhoneg)*(x/X)^kappa);
